package bitRanks

import ap.parser.{Environment, IExpression, ITerm, IFormula, IFunction,
                  IFunApp, CollectingVisitor}
import ap.util.{Debug => ApDebug}

import bitRanks.util.Debug


object OperandSorter {

  private val AC = Debug.AC_OPERAND_SORTER
  
  class OperandSorterException(msg : String) extends Exception(msg)

}

/**
 * Class to traverse expressions and sort the arguments of certain (commutative)
 * binary operators such that the second argument preferable is a small literal.
 * This is useful, because (by convention) operators such as bit-wise and are
 * defined by recursion over the second argument.
 */
class OperandSorter(env : Environment) {
  
  val commOps = {
    val commOps = new scala.collection.mutable.HashSet[IFunction]
    val reader = ResourceFiles.commOpsReader
    
    var cont = true
    while (cont) {
      val line = reader.readLine
      if (line == null) {
        cont = false
      } else {
        (env lookupSym line.trim) match {
          case Environment.Function(fun) => {
            if (fun.arity != 2)
              throw new OperandSorter.OperandSorterException(
                       "Only binary functions can be commutative, not " + fun)
            commOps += fun
          }
          case x =>
            throw new OperandSorter.OperandSorterException(
                     "Only functions can be commutative, not " + x)
        }
      }
    }
    
    reader.close
    
    commOps
  }

  //////////////////////////////////////////////////////////////////////////////

  def apply(t : IExpression) : IExpression =
    SortingVisitor.visit(t, null)
  def apply(t : IFormula) : IFormula =
    SortingVisitor.visit(t, null).asInstanceOf[IFormula]
  
  //////////////////////////////////////////////////////////////////////////////
  
  private object SortingVisitor extends CollectingVisitor[Unit, IExpression] {
    
    import IExpression._
    
    def postVisit(t : IExpression,
                  arg : Unit,
                  subres : Seq[IExpression]) : IExpression = t match {
      case IFunApp(fun,
                   Seq(Const(value1), Const(value2))) if (commOps contains fun) =>
        if (value1.abs < value2.abs)
          fun(value2, value1)
        else
          fun(value1, value2)
      case IFunApp(fun, Seq(Const(value), _)) if (commOps contains fun) =>
        fun(subres(1).asInstanceOf[ITerm], value)
      case _ =>
        t update subres
    }

  }
  
}
